/*
 * F s e t . c
 *
 * This file contains the grammar analysis section of antlr.
 *
 * Terence Parr
 * Purdue University
 * Summer 1989
 *
 * Forced variable names to be unique to 7 char for portability
 * Terence Parr -- March 1990
 */

#include <stdio.h>
#ifdef MEMCHK
#include "trax.h"
#endif
#include "set.h"
#include "struct.h"
#ifdef MPW
/*#pragma segment fset*/
#endif

/*
 * Sometimes the computation of a Follow set will be stopped before completion
 * due to infinite-recursion avoidance checks, namely:
 *		if ( set_el(rule, FoVisit) ) ...Don't Visit rule...
 * This implies that the follow set is incomplete and must be finished later.
 * Two sets are now used for each rule found in a grammar--follow and remainder.
 * The follow set is the current state of the follow computation.  The remainder
 * set is a set of rules (not tokens like the follow) and records the rules that
 * were required for a follow set, but that could not be completed due to infinite
 * recursion.
 *
 * Follow will look into the two tables of sets first to see if a partial solution
 * is known.  If nothing is there, the normal follow computation is initiated.
 * Else, the follow of each rule in the remainder set is attempted and combined
 * with the current follow set.
 *
 * Note that this is very different from the FIRST cache system.  No such complexity
 * is required there.
 *
 * This algorithm modification done January 1990
 * By Terence Parr
 */
static set	*FoCache,		/* Current follow set for a rule */
			*Remainder;		/* Remaining rules needed in follow computation */
static char *FoValid;		/* Is a follow set for a rule valid yet? */

static set **FiCache = NULL;


/*
 * Compute the first set of a rule.
 *
 * Notice that we leave "a trail of bread-crumbs" to detect infinite-recursion
 * (markFi and unMarkFi).
 */
set *
First(rule)
int rule;
{
	set *f;
	set t,u;
	int alt;

	DEBUG(2) {
		fprintf(stderr, "First(%s)\n", RuleStr[rule]);
	}
	f = fBlock(Rules[rule], rule, aRule);

	for (alt=0; f[alt].n != 0; alt++)
	{
		if ( set_el(EpToken, f[alt]) )	/* If optional alternative */
		{
			t = Follow(rule);
			/* Leave Epsilon if follow is empty */
			if ( set_deg(t) != 0 ) set_rm(EpToken, f[alt]);
			u = set_and(f[alt], t); 
			if ( !set_nil(u) )
			{
				warnFLNoCR( eMsg("alt %d rule %s ambiguous",
								alt+1, RuleStr[rule]),
						FileStr[RuleFile[rule]], RuleLine[rule] );
				fprintf(stderr, " upon ");
				if ( set_deg(u) > 1 ) set_fpT(stderr, u);
				else fprintf(stderr, "%s", TokenName(set_int(u)));
				fprintf(stderr, "\n");
			}
			set_orin(&(f[alt]), t);
			set_free( t );
			set_free(u);
		}
	}

	return( f );
}

/*
 * Compute the first set of a block of alternatives.  An array of sets
 * is returned.  set i is the first of alternative i.
 */
set *
fBlock(blk, rule, type)
Block blk;
int rule, type;
{
	set *f;
	int i;
	
	DEBUG(2) fprintf(stderr, "fBlock(%s)\n", RuleStr[rule]);
	/* If Block was cached, use it.  But, only rules are saved for now */
	if ( type == aRule && FiCache[rule] != NULL )	/* If in cache */
	{
		Save++;
		f = dupFset(FiCache[rule]);
	}
	else
	{
		if ( type == aRule ) markFi( rule );
		f = newFset();
		for (i=1; blk[i] != NULL; i++)
		{
			f[i-1] = fAlt(blk[i], rule);
		}
		if ( type == aRule )
		{
			FiCache[rule] = dupFset(f);
			unMarkFi( rule );
		}
	}
	return( f );
}

/*
 * Return the first set of an alternative starting from a
 * specified starting atom --> 'start'.
 *
 * Compute fAtom on each atom starting from 'start' while epsilon is
 * a member of the FIRST set.  In other words, the first of an alternative
 * is the set of all tokens that can possibly begin that alternative either
 * by appearing first or by following optional paths.
 *
 * When optional rule atoms are present, one must make sure that the combined
 * first sets are unique.  This means that one token cannot be accepted by
 * more than one place along the optional paths.
 */
set
fMidAlt(start, rule)
Atom *start;
int rule;
{
	set f, g, u;
	Atom *p;
	
	DEBUG(2) {
		fprintf(stderr, "fMidAlt(%s, @", RuleStr[rule]);
		dispAtom(*start, stderr);
		fprintf(stderr, ")\n");
	}
	p = start;
	skipActions(p);
	if ( p->type == EOL )	/* Epsilon Transfer (empty alternative)? */
	{
		return( set_of(EpToken) );
	}
	f = fAtom(p++, rule, TRUE);
	skipActions(p);
	while ( set_el(EpToken, f) && p->type != EOL )	/* Was atom optional? */
	{
		set_rm(EpToken, f);
		g = fAtom(p, rule, TRUE);
		u = set_and(f, g); 
		if ( !set_nil(u) )
		{
			warnFLNoCR( eMsg("alt of rule %s ambiguous", RuleStr[rule]),
					p->file, p->line );
			fprintf(stderr, " upon ");
			if ( set_deg(u) > 1 ) set_fpT(stderr, u);
			else fprintf(stderr, "%s", TokenName(set_int(u)));
			fprintf(stderr, " near");
			dispAtom( *p, stderr );
			fprintf(stderr, "\n");
		}
		p++;
		set_orin(&f, g);
		set_free(g);
		set_free(u);
		skipActions(p);
	}

	return( f );
}

/*
 * Return the first set of an alternative within a block.
 */
set
fAlt(alt, rule)
Alt alt;
int rule;
{
	return( fMidAlt(&(alt[1]), rule) );
}

/*
 * Compute the first set of an atom.
 * Return the set containing:
 *
 *		one atom			--> If leaf node
 *		First( nonTerm )	--> If atom is rule
 *		fBlock( subBlock )	-->	If atom is a subrule of any type
 *
 * If noAlts, compact f before returning it when subBlock.
 * If !noAlts, a pointer to an array of sets is returned
 *		else a pointer to a single set is returned.
 * If noAlts, then no blk follow is done.
 *
 * noAlts==FALSE only when generating code and you want an array of sets
 * to come back reflecting the FIRST of each alternative within that block.
 */
set
fAtom(atom, rule, noAlts, fset)
Atom *atom;
int rule, noAlts;
set **fset;		/* only used when !noAlts */
{
	int r,n;
	set *f, a;

	DEBUG(2) {
		fprintf(stderr, "fAtom(%s,", RuleStr[rule]);
		dispAtom(*atom, stderr );
		fprintf(stderr, ", %s)\n", noAlts ? "noAlts" : "Alts");
	}
	switch ( atom->type )
	{
		case aLeaf :
			return( set_of(atom->token) );
		case aRule :
			r = ruleNum(atom);
			if ( r == 0 )
			{
				warnFL( eMsg("rule %s undefined", atom->ptr.rule),
						atom->file, atom->line );
				return( empty );
			}
			if ( !set_el(r, FiVisit) )	/* Infinite recursion? */
			{
				f = fBlock(Rules[r], r, aRule);
				a = fCompact(f);
				freeFset(f);
				return( a );
			}
			if ( rule == r )
			{
				warnFL( eMsg("rule %s infinite left-recursion possible", RuleStr[r]),
						atom->file, atom->line );
			}
			else
			{
				warnFL(
					eMsg("rule %s indirect left-recursion to rule %s possible",
						RuleStr[rule], RuleStr[r]),
						atom->file, atom->line );
			}
			return( empty );
		case aSubBLK :
		case aOptBLK :
		case aLoopBLK :
			if ( atom->ptr.block == NULL ) return( set_of(EpToken) );
			if ( emptyBlk(atom->ptr.block) ) return( set_of(EpToken) );
			f = fBlock(atom->ptr.block, rule, atom->type);
			if ( atom->type != aSubBLK )	/* Add implicit Ep Transfer */
			{
				set_orel(EpToken, &(f[nextAlt(f)]));
			}
			if ( (n=multEpsilon(f)) > 1 )
			{
				warnBlk( *atom, eMsg(" [rule %s] has %d epsilon paths",
							RuleStr[rule], n) );
			}
			if ( !noAlts )
			{
				doBlkFollow(f, atom, rule);
				*fset = f;
				return( empty );	/* return value is ignored */
			}
			a = fCompact(f);
			freeFset(f);
			return( a );
	}
	return( empty );
}

/* Compute the follow of a block.  This is similar to the FOLLOW of a rule, but
 * blocks (subrules) can be referenced only in the rule that uses it.
 *
 * Each alternative in the array of sets passed in is checked for optionality
 * (contains epsilon).  If an alternative is optional, the FOLLOW of the subrule
 * computed and appended to the set for that alt.
 */
doBlkFollow(f, atom, rule)
set *f;
Atom *atom;
int rule;
{
	set t,s, fBlkFollow();
	int alt;

	DEBUG(2) {
		fprintf(stderr, "doBlkFollow(%s, @", RuleStr[rule]);
		dispAtom(*atom, stderr);
		fprintf(stderr, ")\n");
	}
	for (alt=0; f[alt].n != 0; alt++)
	{
		if ( set_el(EpToken, f[alt]) )
		{
			set u;
			
			set_rm(EpToken, f[alt]);
			t = fBlkFollow(atom, rule);
			u = set_and(f[alt], t); 
			if ( !set_nil(u) )
			{
				warnBlkNoCR( (*atom), eMsg(" alt %d ambiguous upon ", alt+1) );
				if ( set_deg(u) > 1 ) set_fpT(stderr, u);
				else fprintf(stderr, "%s", TokenName(set_int(u)));
				fprintf(stderr, "\n");
			}
			set_free( u );
			set_orin(&(f[alt]), t);
			/* If what follows can be bypassed til end of rule, do follow */
			/* If no follow, leave Epsilon Token in its place */
			if ( set_el(EpToken, t) )
			{
				s = Follow(rule);
				if ( set_deg(s) > 0 ) set_rm(EpToken, f[alt]);
				u = set_and(f[alt], s);
				if ( !set_nil(u) )
				{
					warnBlkNoCR( (*atom), eMsg(" alt %d ambiguous upon ", alt+1) );
					if ( set_deg(u) > 1 ) set_fpT(stderr, u);
					else fprintf(stderr, "%s", TokenName(set_int(u)));
					fprintf(stderr, "\n");
				}
				set_free( u );
				set_orin(&(f[alt]), s);
				set_free( s );
			}
			set_free( t );
		}
	}
}

/*
 * Do the actual computation for doBlkFollow() above.  This looks at the
 * tokens following the block passed in.
 *
 * When the edge of a subrule is found, the follow of the enclosing block
 * must also be considered.  The parent pointer, points to the subrule atom
 * of the (parent) inner-most enclosing subrule
 */
static set
fBlkFollow(blk, rule)
Atom *blk;
int rule;
{
	set f, t;
	
	DEBUG(2) {
		fprintf(stderr, "fBlkFollow(%s, after:", RuleStr[rule]);
		dispAtom(*blk, stderr);
		fprintf(stderr, ")\n");
	}
	f = fMidAlt(blk+1, rule);
	if ( set_el(EpToken, f) && blk->parent != NULL )
	{
		set_rm(EpToken, f);
		t = fBlkFollow(blk->parent, rule);
		set_orin(&f, t);
		set_free(t);
	}
	return( f );
}

/*
 * Compute the FOLLOW of 'frule'.
 *
 * First check to see if the follow has already been computed.  If so, simply
 * return the cached set.
 *
 * If the FOLLOW has not been totally computed, has it been partially computed?
 * If not, we must perform the FOLLOW--follow().
 *
 * If the FOLLOW set has been partially computed, append the FOLLOW of each
 * rule in the Partial[frule] set to FoCache[frule].  The FOLLOW always
 * be complete after this operation.  This is because the only way for a
 * partial set to arise is when cycles appear in the FOLLOW computation.
 * For example, FOLLOW of rule e requires FOLLOW of rule e2 which requires the
 * FOLLOW of rule e1 which requires FOLLOW of e.  Since we have already begun
 * the FOLLOW computation for rule e, we cannot ask for it again or infinite-
 * recursion will result.  So, the fact that the FOLLOW for rules e1 and e2
 * require FOLLOW rule e is recorded.  After FOLLOW rule e has completed, we
 * have the following information:
 *
 *	Rule	FoCache		Remainder
 *	[e] 	{A B T}
 *	[e1]	{A}			{e}
 *	[e2]	{A B}		{e}
 *
 * FOLLOW for rules e1 and e2 are obviously easy to compute:  Simply append
 * the FOLLOW of rule e to both of them.
 *
 *		a  : I e T s;
 *		e  : e1 {A e1} ;
 *		e1 : e2 {B e2} ; 
 *		e2 : C | D e ;
 */
set
Follow(frule)
int frule;
{
	unsigned *rem, *r;
	
	FoRequest++;				/* We have another follow request */
	DEBUG(2) fprintf(stderr, "Start of Follow(%s)\n", RuleStr[frule]);
	if ( !UseCache )
	{
		follow( frule );
		return( set_dup(FoCache[frule]) );
	}

	if ( FoValid[frule] )		/* If we have fully computed follow */
	{
		FoSave++;				/* Saved complete follow set computation */
		DEBUG(1) {
			fprintf(stderr, "Follow(%s) complete: ", RuleStr[frule]);
			set_fpT(stderr, FoCache[frule]);
			fprintf(stderr, "\n\n");
		}
		return( set_dup(FoCache[frule]) );
	}
	/* If invalid and empty, we have done nothing */
	if ( set_deg( Remainder[frule] ) == 0 )	/* We have to compute whole thing */
	{
		follow( frule );
		set_rm(frule, Remainder[frule]);	/* Make sure no follow of self */
		if ( set_deg(Remainder[frule]) == 0 ) FoValid[frule] = 1;
		DEBUG(1) {
			fprintf(stderr, "(No remainder was avail) after Follow(%s) calc\n", RuleStr[frule]);
			dumpFcache();
		}
		return( set_dup(FoCache[frule]) );
	}

	/* Else we have follows remaining to compute */
	FoPartial++;				/* Partial follow to compute */
	r = rem = set_pdq(Remainder[frule]);
	DEBUG(1) {
		fprintf(stderr,"For Follow(%s), we have these left to compute: ",
						RuleStr[frule]);
		set_fpR(stderr, Remainder[frule]);
		fprintf(stderr, "\n");
	}
	while ( *r != nil )
	{
		if ( !set_el(*r, FoVisit) )
		{
			/*
			 * if *r in remainder set and not in FoVisit this implies
			 * that FOLLOW(*r) is completed and in the cache.
			 */
			/* t = Follow( *r ); set_free( t ); */
			set_orin(&(FoCache[frule]), FoCache[*r]);
			set_orin(&(Remainder[frule]), Remainder[*r]);
			set_rm(*r, Remainder[frule]);
		}
		r++;
	}
	free( rem );
	if ( set_deg(Remainder[frule]) == 0 ) FoValid[frule] = 1;
	DEBUG(1) {
		fprintf(stderr, "(Remainder was avail) after Follow(%s) calc\n",
						RuleStr[frule]);
		dumpFcache();
	}
	return( set_dup(FoCache[frule]) );
}

/*
 * Include Epsilon if follow required and is empty
 * Do not do followBlk for rules that do not reference frule
*/
static
follow(frule)
int frule;
{
	int r;
	set t;
	
	FoCompute++;
	markFo( frule );
	for (r=1; Rules[r]!=NULL; r++)
	{
		if ( !set_el(r, CrossRef[frule]) ) continue;
		t = followBlk(Rules[r], frule, r, NULL);
		set_orin(&(FoCache[frule]), t);
		set_free(t);
		if ( set_el(EpToken, FoCache[frule]) )
		{
			set_rm(EpToken, FoCache[frule]);
			if ( set_el(r, FoVisit) )
			{
				set_orel( r, &(Remainder[frule]) );
			}
			else
			{
				t = Follow( r );
				set_free( t );
				set_orin(&(FoCache[frule]), FoCache[r]);
				set_orin(&(Remainder[frule]), Remainder[r]);
				if ( set_deg(FoCache[frule]) == 0 )
				{
					set_orel(EpToken, &(FoCache[frule]));
				}
			}
		}
	}
	unMarkFo( frule );
}

/*
 * We know that rule 'frule' is referenced in 'curRule' because of
 * CrossRef[].  Check this block for a reference to it.
 */
set
followBlk(blk,frule,curRule,parent)
Block blk;
Atom *parent;
int frule, curRule;
{
	int alt;
	set t, f;
	
	f = empty;
	for (alt=1; blk[alt]!=NULL; alt++)
	{
		t = followAlt(blk[alt], frule, curRule);
		set_orin(&(f), t);
		set_free( t );
		DEBUG(3) setMsg( eMsg("followAlt(%s) %d returns ",RuleStr[frule],alt), f );
	}
	if ( parent != NULL )	/* No parent => can't do fBlkFollow() */
	{
		if ( set_el(EpToken, f ) )	/* Follow past end of block? */
		{
			if ( parent->type == aLoopBLK )
			{
				/* Must include first of loop block ()* itself */
				t = fMidAlt(parent, curRule);
				set_orin(&(f), t);
				set_free( t );
			}
			set_rm(EpToken, f);
			t = fBlkFollow(parent, curRule);
			set_orin(&(f), t);
			set_free( t );
		}
	}
	DEBUG(3) setMsg( eMsg("followBlk(%s) returns",RuleStr[frule]), f );
	return( f );
}

/*
 * For each atom in an alternative, check subrules for references to 'frule'.
 * Also, if atom is a ref to 'frule', append FIRST of alt[atom+1] to the
 * FOLLOW of 'frule'.
 */
set
followAlt(alt,frule,curRule)
Atom *alt;
int frule, curRule;
{
	set t, f;
	int a,r;

	DEBUG(2) {
		fprintf(stderr,"followAlt(%s in %s @",RuleStr[frule],RuleStr[curRule]);
		dispAtom( alt[1], stderr );
		fprintf(stderr, ")\n");
	}
	f = empty;
	for (a=1; alt[a].type!=EOL; a++)
	{
		switch ( alt[a].type )
		{
			case aLeaf :
			case aAction :
				continue;
			case aSubBLK :
			case aOptBLK :
			case aLoopBLK :
				t = followBlk(alt[a].ptr.block,frule,curRule,&(alt[a]));
				set_orin(&(f), t);
				set_free( t );
				continue;
			case aRule :
				r = ruleNum(&(alt[a]));
				if ( r == 0 )
				{
					warnFL(eMsg("rule %s undefined", alt[a].ptr.rule),
							alt[a].file, alt[a].line);
					continue;
				}
				if ( frule != r ) continue;
				/* Found match. "following" rule is in this alternative */
				t = fMidAlt(&(alt[a+1]), curRule);
				set_orin(&(f), t);
				set_free( t );
				continue;
		}
	}
	return( f );
}

/* return number of epsilon paths in array of sets */
multEpsilon( f )
set *f;
{
	int found = 0;
	
	while ( f->n != 0 )
	{
		if ( set_el(EpToken, *f++) ) found++;
	}
	return( found );
}

ruleNum(a)
register Atom *a;
{
	if ( a->token != 0 || RuleStr[1] == 0 ) return( a->token );
	return( (a->token=strPos(RuleStr, a->ptr.rule)) );
}

/*
 * Compact f and place into set a.  i.e. a = OR(f[0]|..|f[n-1])
 */
set
fCompact( f )
set *f;
{
	set a;
	
	a = empty;
	while ( f->n != 0 )
	{
		set_orin(&a, *f++);
	}
	return( a );
}

/* return TRUE if block is empty (no rule elements other than actions */
emptyBlk(b)
Atom **b;
{
	Atom *p;
	int i=1;
	
	if ( b == NULL ) return(1);
	b++;	/* Move to 1st element in block */
	while ( *b != NULL )
	{
		p = (*b)+1;
		while ( p->type != EOL )
		{
			if (p->type != aAction) return(0);
			p++;
		}
		b++;
	}
	return(1);
}

fDump( f )
set *f;
{
	int i=0;

	while ( f->n != 0 )
	{
		setMsg( eMsg("[%d] ", i), *f);
		i++;
		f++;
	}
}

fset_pr(e)
set *e;
{
	unsigned i=0;

	while ( e->n != 0 )
	{
		printf("[%d] ", i++);
		set_prT(*e++);
		putchar('\n');
	}
}

set_fpR(f, e)
FILE *f;
set e;
{
	register unsigned *p;
	unsigned *q;

	if ( set_deg(e) == 0 ) return;
	if ( (q=p=set_pdq(e)) == NULL ) fatal("Can't alloc space for set_pdq");
	fprintf(f, "{");
	while ( *p != nil )
	{
		fprintf(f, " %s", RuleStr[*p++]);
	}
	fprintf(f, " }");
	free(q);
}

set_fpT(f, e)
FILE *f;
set e;
{
	register unsigned *p;
	unsigned *q;

	if ( set_deg(e) == 0 ) return;
	if ( (q=p=set_pdq(e)) == NULL ) fatal("Can't alloc space for set_pdq");
	fprintf(f, "{");
	while ( *p != nil )
	{
		fprintf(f, " %s", TokenName(*p++));
	}
	fprintf(f, " }");
	free(q);
}

set_prR(e)
set e;
{
	set_fpR(stdout, e);
}

set_prT(e)
set e;
{
	set_fpT(stdout, e);
}

set *
newFset()
{
	set *f;
	int i;
	
	f = (set *) calloc(MaxAlts+1, sizeof(set));
	if ( f==NULL ) fatal("Cannot allocate new fset\n");
	for (i=0; i<MaxAlts+1; i++) f[i] = empty;
	return( f );
}

/*
 * Duplicate an array of sets.  Return a pointer to it.
 */
set *
dupFset(f)
set *f;
{
	set *g, *p;
	
	p = g = newFset();
	while ( f->n != 0 ) *p++ = set_dup( *f++ );
	return( g );
}

void
freeFset(f)
set *f;
{
	set *p=f;
	
	if ( f == 0 ) return;
	while ( p->n != 0 )
	{
		set_free( *p );
		p++;
	}
	free( f );
}

/*
 * return the next empty set found in f.  f must hold an empty one or
 * this routine will fall off the edge of the world.
 */
nextAlt(f)
set *f;
{
	int i=0;
	
	while ( f->n != 0 ) {f++; i++;}
	return( i );
}

/* C a c h e  R o u t i n e s */

InitCache(n)
int n;			/* Number of rules */
{
	if ( n <= 0 ) return;
	FiCache = (set **) calloc(n, sizeof(set *));
	if ( FiCache == NULL ) fatal("Cannot allocate FIRST cache\n");
	FoCache = (set *) calloc(n, sizeof(set));
	if ( FoCache == NULL ) fatal("Cannot allocate FoCache\n");
	Remainder = (set *) calloc(n, sizeof(set));
	if ( Remainder == NULL ) fatal("Cannot allocate Remainder\n");
	FoValid = (char *) calloc(n, sizeof(char));
	if ( FoValid == NULL ) fatal("Cannot allocate FoValid\n");
}

KillCache(n)
int n;
{
	int r;

	if ( n <= 0 ) return;
	if ( FiCache != NULL )
	{
		register set **p;
		
		p = &(FiCache[n-1]);
		while ( p >= &(FiCache[1]) )
		{
			if ( *p != NULL ) freeFset( *p );
			--p;
		}
		free( FiCache );
	}
	if ( FoCache != NULL )
	{
		for (r=1; r<=NumRules; r++)
		{
			if ( FoCache[r].n != 0 ) set_free( FoCache[r] );
		}
		free( FoCache );
	}
	if ( Remainder != NULL )
	{
		for (r=1; r<=NumRules; r++)
		{
			if ( FoCache[r].n != 0 ) set_free( Remainder[r] );
		}
		free( Remainder );
	}
	if ( FoValid != NULL ) free( FoValid );
}

dumpFcache()
{
	int r;
	
	fprintf(stderr, "Follow cache state:\n");
	for (r=1; r<=NumRules; r++)
	{
		fprintf(stderr, "Follow[%s] = ", RuleStr[r]);
		set_fpT(stderr, FoCache[r]);
		if ( !FoValid[r] ) fprintf(stderr, "\txxx\n");
		else fprintf(stderr, "\n");
	}
	for (r=1; r<=NumRules; r++)
	{
		fprintf(stderr, "Remainder[%s] = ", RuleStr[r]);
		set_fpR(stderr, Remainder[r]);
		fprintf(stderr, "\n");
	}
	fprintf(stderr, "\n\n");
}

